


// GRExampleHandler1.cs
// 2007 July 31; Colin Fahey (colinfahey.com)




using System;
using System.Collections.Generic;
using System.Text;
using System.IO;
using System.Windows.Forms;

using GRV11; // Requires GRV11.cs or a "reference" to GRV11DLL.dll

using System.Drawing;
using System.Drawing.Imaging;





public class GRExampleHandler1
{
  private int mLargeFontOpenGLDisplayListBase;
  private int mSmallFontOpenGLDisplayListBase;
  private GRMesh mGRMesh;
  private GRTexture mGRTexture;
  private GRVector3 mCameraPosition;
  private int mShaderProgramHandle = 0;

  private Bitmap mHUDBitmap;
  private GRTexture mHUDGRTexture;




  public GRExampleHandler1()
  {
    this.mLargeFontOpenGLDisplayListBase = 0;
    this.mSmallFontOpenGLDisplayListBase = 0;
    this.mGRMesh = new GRMesh();
    this.mGRTexture = new GRTexture();
    this.mCameraPosition.Set(0.0f, 0.0f, 600.0f);
    this.mShaderProgramHandle = 0;

    this.mHUDBitmap = new Bitmap(1024, 128, System.Drawing.Imaging.PixelFormat.Format32bppArgb);
    this.mHUDGRTexture = new GRTexture();
  }




  public static String WoodVertexShaderSource =
  @"
// wood.vert
//
// Simple vertex shader for wood
//
// Author: John Kessenich
//
// Copyright (c) 2002-2004 3Dlabs Inc. Ltd.
//
// See 3Dlabs-License.txt for license information
//

varying float lightIntensity;
varying vec3 Position;

uniform vec3 LightPosition;
uniform float Scale;

void main(void)
{
    vec4 pos = gl_ModelViewMatrix * gl_Vertex;
    Position = vec3(gl_Vertex) * Scale;
    vec3 tnorm = normalize(gl_NormalMatrix * gl_Normal);
    lightIntensity = max(dot(normalize(LightPosition - vec3(pos)), tnorm), 0.0) * 1.5;
    gl_Position = gl_ModelViewProjectionMatrix * gl_Vertex;
}
";




  public static String WoodFragmentShaderSource =
    @"//
// Simple fragment shader for wood
//
// Author: John Kessenich
//
// Copyright (c) 2002-2004 3Dlabs Inc. Ltd.
//
// See 3Dlabs-License.txt for license information
//

uniform float GrainSizeRecip;
uniform vec3  DarkColor;
uniform vec3  spread;

varying float lightIntensity; 
varying vec3 Position;

void main (void)
{
    //
    // cheap noise
    //
    vec3 scaledPosition = Position * 0.001;
    vec3 location = scaledPosition;
    vec3 floorvec = vec3(floor(10.0 * scaledPosition.x), 0.0, floor(10.0 * scaledPosition.z));
    vec3 noise = scaledPosition * 10.0 - floorvec - 0.5;
    noise *= noise;
    location += noise * 0.12;

    //
    // distance from axis
    //
    float dist = location.x * location.x + location.z * location.z;
    float grain = dist * GrainSizeRecip;

    //
    // grain effects as function of distance
    //
    float brightness = fract(grain);
    if (brightness > 0.5) 
        brightness = (1.0 - brightness);
    vec3 color = DarkColor + brightness * spread;
    
    brightness = fract(grain * 7.0);    
    if (brightness > 0.5) 
        brightness = 1.0 - brightness;
    color -= brightness * spread;

    //
    // also as a function of lines parallel to the axis
    //
    brightness = fract(grain * 47.0) * 0.60;
    float line = fract(scaledPosition.z + scaledPosition.x);
    float snap = floor(line * 20.0) * (1.0/20.0);
    if (line < snap + 0.006)
        color -= brightness * spread;

    //
    // apply lighting effects from vertex processor
    //
    color = clamp(color * lightIntensity, 0.0, 1.0); 
    gl_FragColor = vec4(color, 1.0);
}";




  public static String ToonVertexShaderSource =
  @"
// toon.vert : simple toon vertex shader

varying vec3 normal, lightDir;

void main()
{
    lightDir = normalize(vec3(gl_LightSource[0].position));
    normal = normalize(gl_NormalMatrix * gl_Normal);

    gl_Position = ftransform();
}
";



  public static String ToonFragmentShaderSource =
  @"
// toon.frag : simple toon fragment shader

varying vec3 normal, lightDir;

void main()
{
    float intensity;
    vec3 n;
    vec4 color;

    n = normalize(normal);

    intensity = max(dot(lightDir,n),0.0);
    if (intensity > 0.98)
      color = vec4(0.8,0.8,0.8,1.0);
    else if (intensity > 0.5)
      color = vec4(0.4,0.4,0.8,1.0);
    else if (intensity > 0.25)
      color = vec4(0.2,0.2,0.4,1.0);
    else
      color = vec4(0.1,0.1,0.1,1.0);

    gl_FragColor = color;
}
";





  public void OpenGLStarted(GRControl grControl)
  {
    GR gr = grControl.GetGR();

    if (null == gr) return;

    GRSupport.SupportCreateBitmapFont
    (
        gr,
        grControl.GetHDC(),
        "Verdana",
        14,
        ref this.mLargeFontOpenGLDisplayListBase
    );

    GRSupport.SupportCreateBitmapFont
    (
        gr,
        grControl.GetHDC(),
        "Verdana",
        10,
        ref this.mSmallFontOpenGLDisplayListBase
    );


    // Load shader and set variables
    int shaderChoice = 1;
    if (true == gr.bglCreateProgramObjectARB)
    {
      if (1 == shaderChoice)
      {
        GRSupport.SupportShaderProgramCreate
        (
            gr,
            GRExampleHandler1.WoodVertexShaderSource,
            GRExampleHandler1.WoodFragmentShaderSource,
            ref this.mShaderProgramHandle
        );

        // Start using the program immediately so we can set variables
        GRSupport.SupportShaderProgramSelect(gr, this.mShaderProgramHandle);

        int shaderVarf3LightPosition = 0;
        int shaderVarfGrainSizeRecip = 0;
        int shaderVarfScale = 0;
        int shaderVarf3DarkColor = 0;
        int shaderVarf3spread = 0;

        shaderVarf3LightPosition = (int)gr.glGetUniformLocationARB(this.mShaderProgramHandle, "LightPosition");
        shaderVarfGrainSizeRecip = (int)gr.glGetUniformLocationARB(this.mShaderProgramHandle, "GrainSizeRecip");
        shaderVarfScale = (int)gr.glGetUniformLocationARB(this.mShaderProgramHandle, "Scale");
        shaderVarf3DarkColor = (int)gr.glGetUniformLocationARB(this.mShaderProgramHandle, "DarkColor");
        shaderVarf3spread = (int)gr.glGetUniformLocationARB(this.mShaderProgramHandle, "spread");

        gr.glUniform3fARB(shaderVarf3LightPosition, 400.0f, 400.0f, 400.0f);
        gr.glUniform1fARB(shaderVarfGrainSizeRecip, 4.0f);
        gr.glUniform1fARB(shaderVarfScale, 10.0f);
        gr.glUniform3fARB(shaderVarf3DarkColor, 0.6f, 0.3f, 0.1f);
        gr.glUniform3fARB(shaderVarf3spread, 0.15f, 0.075f, 0.0f);
      }
      else if (2 == shaderChoice)
      {
        GRSupport.SupportShaderProgramCreate
        (
            gr,
            GRExampleHandler1.ToonVertexShaderSource,
            GRExampleHandler1.ToonFragmentShaderSource,
            ref this.mShaderProgramHandle
        );
      }
    }


    this.mGRMesh = new GRMesh();
    this.mGRMesh.BuildSimpleCube(200.0f);

    if (false == gr.bglCreateProgramObjectARB)
    {
      this.mGRTexture = new GRTexture();
      this.mGRTexture.CreateCheckerboardTexture(gr, 256, 256);
    }



    this.mHUDGRTexture.CreateTextureFromBitmap(gr, this.mHUDBitmap, true);

    // To prevent "tearing" (irregular streaks) due to swapping buffers at
    // arbitrary times relative to the vsync times, we indicate that we wish 
    // to wait for vsync before swapping buffers.
    // This request applies when the display driver control panel -- in the
    // OpenGL settings area -- is set to let the application decide whether or
    // not to wait for vsync.  Otherwise, the driver control panel overrides
    // any request made here.
    if (true == gr.bwglSwapIntervalEXT)
    {
      gr.wglSwapIntervalEXT(1);
    }

  }





  public void Paint(object sender, PaintEventArgs e)
  {
    if (null == sender) return;
    if (false == (sender is GRControl)) return;

    GRControl grControl = (sender as GRControl);
    GR gr = grControl.GetGR();

    int clientWidth = grControl.ClientRectangle.Width;
    int clientHeight = grControl.ClientRectangle.Height;
    if (clientWidth <= 0)
    {
      clientWidth = 1;
    }
    if (clientHeight <= 0)
    {
      clientHeight = 1;
    }


    // Viewport 
    gr.glViewport(0, 0, clientWidth, clientHeight);


    // Clear the viewport
    gr.glClearColor(0.5f, 0.5f, 0.5f, 1.0f);
    gr.glClear(GR.GL_COLOR_BUFFER_BIT | GR.GL_DEPTH_BUFFER_BIT);


    // Basic rendering conditions
    gr.glEnable(GR.GL_DEPTH_TEST);
    gr.glDepthFunc(GR.GL_LEQUAL);
    gr.glEnable(GR.GL_CULL_FACE);
    gr.glCullFace(GR.GL_BACK);
    gr.glFrontFace(GR.GL_CCW);


    // PROJECTION matrix, typically for perspective correction or orthographic projection
    gr.glMatrixMode(GR.GL_PROJECTION);
    gr.glLoadIdentity();
    double aspectRatio = 1.0;
    if (0 != clientHeight)
    {
      aspectRatio = ((double)(clientWidth) / (double)(clientHeight));
    }
    double verticalFieldOfViewAngle = 60.0;
    gr.gluPerspective
    (
        verticalFieldOfViewAngle, // Field of view angle (Y angle; degrees)
        aspectRatio, // width/height
        0.1, // distance to near clipping plane
        64000.0 // distance to far clipping plane
    );




    // MODELVIEW matrix, typically used to transform individual models
    gr.glMatrixMode(GR.GL_MODELVIEW);
    gr.glLoadIdentity();

    // Preserve current matrix for the active matrix stack (in this case the MODELVIEW matrix)
    gr.glPushMatrix();

    gr.glTranslatef
    (
        (-(this.mCameraPosition.mX)),
        (-(this.mCameraPosition.mY)),
        (-(this.mCameraPosition.mZ))
    );


    float phase = (float)(30.0 * grControl.GetTotalElapsedTimeSeconds());
    gr.glRotatef((0.11f * phase), 1.0f, 0.0f, 0.0f);
    gr.glRotatef((0.31f * phase), 0.0f, 1.0f, 0.0f);
    gr.glRotatef((0.19f * phase), 0.0f, 0.0f, 1.0f);

    // If shaders are supported, use the shader...
    if (true == gr.bglUseProgramObjectARB)
    {
      GRSupport.SupportShaderProgramSelect(gr, this.mShaderProgramHandle);
    }
    else // ...otherwise use the texture
    {
      if (null != this.mGRTexture)
      {
        this.mGRTexture.SetAsActiveTexture(gr);
      }
    }


    // Draw model(s), using active texture or shader
    this.mGRMesh.Draw(gr);


    // If we used a shader, disable it now...
    if (true == gr.bglUseProgramObjectARB)
    {
      GRSupport.SupportShaderProgramSelect(gr, 0);
    }
    else // ...otherwise we used a texture and we disable it now.
    {
      if (null != this.mGRTexture)
      {
        // Disable texture
        gr.glBindTexture(GR.GL_TEXTURE_2D, 0);
      }
    }


    // Restore the previously-preserved matrix for the active matrix stack (in this case the MODELVIEW matrix)
    gr.glPopMatrix();




    // Draw text 
    gr.glColor4f(0.0f, 0.0f, 0.0f, 1.0f);

    String text = "";


    text = "frame: ";
    text += String.Format("{0}", grControl.GetTotalFramesDrawn());
    text += "  time: ";
    text += String.Format("{0:f3}", grControl.GetTotalElapsedTimeSeconds());

    GRSupport.SupportDrawTextUsingPolygonFont
    (
        gr,
        this.mSmallFontOpenGLDisplayListBase,
        grControl.ClientSize.Width,
        grControl.ClientSize.Height,
        16, // fontHeight
        4, // textX
        4, // textY
        text
    );


    text = "C# wrapper for OpenGL";

    GRSupport.SupportDrawTextUsingPolygonFont
    (
        gr,
        this.mLargeFontOpenGLDisplayListBase,
        grControl.ClientSize.Width,
        grControl.ClientSize.Height,
        20, // fontHeight
        4, // textX
        (grControl.ClientSize.Height - 20), // textY
        text
    );


    //text = "colinfahey.com";

    //GRSupport.SupportDrawTextUsingPolygonFont
    //(
    //    gr,
    //    this.mLargeFontOpenGLDisplayListBase,
    //    grControl.ClientSize.Width,
    //    grControl.ClientSize.Height,
    //    20, // fontHeight
    //    4, // textX
    //    (grControl.ClientSize.Height - 20 - 20), // textY
    //    text
    //);





    // IF YOU WANT TO EXPERIMENT WITH DYNAMIC UPDATING OF TEXTURES WITH GDI+ BITMAPS,
    // USE THE "UNSAFE" COMPILE SETTING, AND UNCOMMENT THE LINE NEAR THE TOP OF THE
    // WRAPPER SOURCE FILE TO ENABLE "UNSAFE" OPTIMIZATIONS.  "UNSAFE" CODE IS 
    // **ESSENTIAL** TO MAKING MEMORY BUFFER COPYING FAST.
    //
    // IF YOU REFER TO "GRV11DLL.dll", THEN IT IS NOT NECESSARY TO COMPILE THIS 
    // APPLICTION WITH "UNSAFE" SET.

#if false
    {
      text = "frame: ";
      text += String.Format("{0}", grControl.GetTotalFramesDrawn());
      text += "  time: ";
      text += String.Format("{0:f3}", grControl.GetTotalElapsedTimeSeconds());

      using (Graphics g = Graphics.FromImage(this.mHUDBitmap))
      {
        g.Clear(Color.FromArgb(0, Color.White));
        g.DrawString(text, new Font("Verdana", 50.0f), Brushes.Black, new PointF(0.0f, 0.0f));
      }
      this.mHUDGRTexture.UpdateTextureWithBitmapData(gr, this.mHUDBitmap);

      GRSupport.SupportDrawTextureImageUnrotatedAndOrthographically
         (
         gr,
         grControl.ClientSize.Width,
           grControl.ClientSize.Height,
         this.mHUDGRTexture,
         0,
         0, // i.e., 0 == draw TOP of image at TOP of viewport, Y-axis points DOWN
         this.mHUDGRTexture.GetWidth(),  // grControl.ClientSize.Width, // this.mHUDGRTexture.GetWidth(),
         this.mHUDGRTexture.GetHeight() // grControl.ClientSize.Height // this.mHUDGRTexture.GetHeight()
         );
    }
#endif



    // Flush all the current rendering and flip the back buffer to the front.
    gr.wglSwapBuffers(grControl.GetHDC());
  }





  public void KeyDown(object sender, KeyEventArgs e)
  {
    if (null == sender) return;
    if (false == (sender is GRControl)) return;

    GRControl grControl = (sender as GRControl);
    GR gr = grControl.GetGR();

    if (e.KeyCode == Keys.A)
    {
      this.mCameraPosition.mY += 10.0f;
    }
    if (e.KeyCode == Keys.Z)
    {
      this.mCameraPosition.mY -= 10.0f;
    }


    // NOTE: The only way for cursor key events (up,down,left,right)
    // to make it to this function is for the main form to implement 
    // the following:
    //
    //   protected override bool ProcessDialogKey ( Keys keyData )
    //
    // and explicitly invoke this KeyDown() method with the 
    // an appropriately formed KeyEventArgs instance.

    if (e.KeyCode == Keys.Up)
    {
      this.mCameraPosition.mZ -= 10.0f;
    }
    if (e.KeyCode == Keys.Down)
    {
      this.mCameraPosition.mZ += 10.0f;
    }
    if (e.KeyCode == Keys.Left)
    {
      this.mCameraPosition.mX -= 10.0f;
    }
    if (e.KeyCode == Keys.Right)
    {
      this.mCameraPosition.mX += 10.0f;
    }


    // Save an image of the viewport (press Shift-0 (zero)).  The following
    // code writes out the viewport in the following image formats: BMP, PNG, GIF, JPG.

    // If you only want a single format, comment out the other file write commands.
    // BMP has no compression artifacts, but the file can be quite large.
    // PNG looks good, and supports 8-bit transparancy (good for textures, etc).
    // GIF looks bad unless you build the color table intelligently (there is a 
    //    neural network color table builder for GIF, in C#/.NET, that you can
    //    find on the Internet; perhaps Paint.NET uses that code); but GIF files
    //    can be quite small, and supports animation.
    // JPG looks good under most circumstances, and the file size can be quite small,
    //    but transparency is not supported.
    // So, for pixel-perfect images, where file size is not important, BMP might be appropriate.
    // For textures with transparency, PNG might be appropriate.
    // For good-looking images, and small file size, and use in Web pages, JPG might be appropriate.
    // For some purposes, with small file sizes, and use in Web pages, GIF might be appropriate.

    // NOTE: IF YOU WANT TO MAKE WRITE OUT IMAGE FILES **SIGNIFICANTLY** FASTER,
    // USE THE "UNSAFE" COMPILE SETTING, AND UNCOMMENT THE LINE NEAR THE TOP OF THE
    // WRAPPER SOURCE FILE TO ENABLE "UNSAFE" OPTIMIZATIONS.  "UNSAFE" CODE IS 
    // **ESSENTIAL** TO MAKING MEMORY BUFFER COPYING FAST.
    //
    // IF YOU REFER TO "GRV11DLL.dll", THEN IT IS NOT NECESSARY TO COMPILE THIS 
    // APPLICTION WITH "UNSAFE" SET.

    if ((e.KeyCode == Keys.D0) && (e.Shift == true))
    {
      DateTime now = DateTime.Now;
      String dateTimeString = String.Format("{0:d4}{1:d2}{2:d2}{3:d2}{4:d2}{5:d2}{6:d3}", now.Year, now.Month, now.Day, now.Hour, now.Minute, now.Second, now.Millisecond);

      String frameIndexString = String.Format("{0:d6}", grControl.GetTotalFramesDrawn());

      String fileNameWithoutExtension = "screen" + "_" + dateTimeString + "_" + frameIndexString;

      GRSupport.SupportWriteViewportToImageFile(gr, fileNameWithoutExtension + ".bmp", System.Drawing.Imaging.ImageFormat.Bmp);
      GRSupport.SupportWriteViewportToImageFile(gr, fileNameWithoutExtension + ".png", System.Drawing.Imaging.ImageFormat.Png);
      GRSupport.SupportWriteViewportToImageFile(gr, fileNameWithoutExtension + ".gif", System.Drawing.Imaging.ImageFormat.Gif);
      GRSupport.SupportWriteViewportToImageFile(gr, fileNameWithoutExtension + ".jpg", System.Drawing.Imaging.ImageFormat.Jpeg);
    }

  }




  public void KeyUp(object sender, KeyEventArgs e)
  {
  }




  private Point mouseClientPositionStart;
  private GRVector3 cameraPositionStart;




  public void MouseDown(object sender, MouseEventArgs e)
  {
    if (null == sender) return;
    if (false == (sender is GRControl)) return;
    GRControl grControl = (sender as GRControl);

    mouseClientPositionStart = grControl.PointToClient(Cursor.Position);
    cameraPositionStart = this.mCameraPosition;

    if ((e.Button & MouseButtons.Left) == MouseButtons.Left)
    {
    }

    if ((e.Button & MouseButtons.Right) == MouseButtons.Right)
    {
      this.mCameraPosition.mZ += 100.0f;
    }
  }



  public void MouseUp(object sender, MouseEventArgs e)
  {
    if ((e.Button & MouseButtons.Right) == MouseButtons.Right)
    {
      this.mCameraPosition.mZ -= 100.0f;
    }
  }



  public void MouseMove(object sender, MouseEventArgs e)
  {
    if (null == sender) return;
    if (false == (sender is GRControl)) return;
    GRControl grControl = (sender as GRControl);

    Point mouseClientPositionCurrent = grControl.PointToClient(Cursor.Position);

    if ((e.Button & MouseButtons.Left) == MouseButtons.Left)
    {
      this.mCameraPosition.mX = this.cameraPositionStart.mX - 0.5f * (float)(mouseClientPositionCurrent.X - mouseClientPositionStart.X);
      this.mCameraPosition.mY = this.cameraPositionStart.mY + 0.5f * (float)(mouseClientPositionCurrent.Y - mouseClientPositionStart.Y);
    }
  }



  public void MouseWheel(object sender, MouseEventArgs e)
  {
    this.mCameraPosition.mZ -= 0.25f * (float)e.Delta;
  }



}

